<?php
/**
 * Horde_Pdf package
 *
 * @license  http://www.horde.org/licenses/lgpl21
 * @category Horde
 * @package  Pdf
 */

/**
 * Exception class for the Horde_Pdf package.  All exceptions thrown
 * from the package will be of this type.
 *
 * @category Horde
 * @package  Pdf
 */
class Horde_Pdf_Exception extends Horde_Exception_Wrapped
{
}
